@@ -21,18 +21,16 @@ use crate::traits::{
 };
 
 pub mod big_int;
-mod digits;
+mod contracts;
 pub mod fraction;
 pub mod traits;
-mod utils;
 
 #[cfg(target_arch = "x86")]
 type Digit = u16;
 #[cfg(not(target_arch = "x86"))]
 type Digit = u32;
-
-const BINARY_SHIFT: usize = (traits::OppositionOf::<Digit>::BITS - 2) as usize;
-const UNDEFINED_DIVISION_ERROR_MESSAGE: &str = "Division by zero is undefined.";
+const BINARY_SHIFT: usize = (Digit::BITS - 1) as usize;
+const _: () = assert!(big_int::is_valid_shift::<Digit, BINARY_SHIFT>());
 const PICKLE_SERIALIZATION_ENDIANNESS: Endianness = Endianness::Little;
 
 type BigInt = big_int::BigInt<Digit, '_', BINARY_SHIFT>;
@@ -126,10 +124,10 @@ impl PyInt {
                     Ok(PyInt(
                         BigInt::try_from(value.extract::<&PyFloat>()?.value()).map_err(
                             |reason| match reason {
-                                big_int::FromFloatConversionError::Infinity => {
+                                big_int::TryFromFloatError::Infinity => {
                                     PyOverflowError::new_err(reason.to_string())
                                 }
-                                big_int::FromFloatConversionError::NaN => {
+                                big_int::TryFromFloatError::NaN => {
                                     PyValueError::new_err(reason.to_string())
                                 }
                             },
@@ -156,8 +154,8 @@ impl PyInt {
     }
 
     #[pyo3(text_signature = "($self, other, /)")]
-    fn gcd(&self, other: Self) -> PyInt {
-        PyInt(self.0.clone().gcd(other.0))
+    fn gcd(&self, other: &Self) -> PyInt {
+        PyInt((&self.0).gcd(&other.0))
     }
 
     #[getter]
@@ -171,13 +169,13 @@ impl PyInt {
     }
 
     fn __abs__(&self) -> PyInt {
-        PyInt(self.0.clone().abs())
+        PyInt((&self.0).abs())
     }
 
     fn __add__(&self, other: &PyAny) -> PyResult<PyObject> {
         let py = other.py();
         if other.is_instance(PyInt::type_object(py))? {
-            Ok(PyInt(self.0.clone() + other.extract::<PyInt>()?.0).into_py(py))
+            Ok(PyInt(&self.0 + other.extract::<PyInt>()?.0).into_py(py))
         } else {
             self.__radd__(other)
         }
@@ -186,7 +184,7 @@ impl PyInt {
     fn __and__(&self, other: &PyAny) -> PyResult<PyObject> {
         let py = other.py();
         if other.is_instance(PyInt::type_object(py))? {
-            Ok(PyInt(self.0.clone() & other.extract::<PyInt>()?.0).into_py(py))
+            Ok(PyInt(&self.0 & other.extract::<PyInt>()?.0).into_py(py))
         } else {
             self.__rand__(other)
         }
@@ -203,7 +201,7 @@ impl PyInt {
     fn __divmod__(&self, divisor: &PyAny) -> PyResult<PyObject> {
         let py = divisor.py();
         match try_py_any_to_maybe_big_int(divisor)? {
-            Some(divisor) => try_divmod(self.0.clone(), divisor)
+            Some(divisor) => try_divmod(&self.0, divisor)
                 .map(|(quotient, remainder)| (PyInt(quotient), PyInt(remainder)).into_py(py)),
             None => Ok(py.NotImplemented()),
         }
@@ -224,7 +222,7 @@ impl PyInt {
         let py = divisor.py();
         match try_py_any_to_maybe_big_int(divisor)? {
             Some(divisor) => {
-                try_floordiv(self.0.clone(), divisor).map(|result| PyInt(result).into_py(py))
+                try_floordiv(&self.0, &divisor).map(|result| PyInt(result).into_py(py))
             }
             None => Ok(py.NotImplemented()),
         }
@@ -243,7 +241,7 @@ impl PyInt {
     }
 
     fn __invert__(&self) -> PyInt {
-        PyInt(!self.0.clone())
+        PyInt(!&self.0)
     }
 
     fn __lshift__(&self, shift: &PyAny) -> PyResult<PyObject> {
@@ -269,14 +267,14 @@ impl PyInt {
     fn __mul__(&self, other: &PyAny) -> PyResult<PyObject> {
         let py = other.py();
         if other.is_instance(PyInt::type_object(py))? {
-            Ok(PyInt(self.0.clone() * other.extract::<PyInt>()?.0).into_py(py))
+            Ok(PyInt(&self.0 * other.extract::<PyInt>()?.0).into_py(py))
         } else {
             self.__rmul__(other)
         }
     }
 
     fn __neg__(&self) -> PyInt {
-        PyInt(-self.0.clone())
+        PyInt(-&self.0)
     }
 
     fn __or__(&self, other: &PyAny) -> PyResult<PyObject> {
@@ -297,10 +295,10 @@ impl PyInt {
         match try_py_any_to_maybe_big_int(exponent)? {
             Some(exponent) => match divisor {
                 Some(divisor) => match try_py_any_to_maybe_big_int(divisor)? {
-                    Some(divisor) => try_pow_mod(self.0.clone(), exponent, divisor, py),
+                    Some(divisor) => try_pow_mod(&self.0, exponent, divisor, py),
                     None => Ok(py.NotImplemented()),
                 },
-                None => try_pow(self.0.clone(), exponent, py),
+                None => try_pow(&self.0, &exponent, py),
             },
             None => Ok(py.NotImplemented()),
         }
@@ -309,7 +307,7 @@ impl PyInt {
     fn __radd__(&self, other: &PyAny) -> PyResult<PyObject> {
         let py = other.py();
         if other.is_instance(PyLong::type_object(py))? {
-            Ok(PyInt(self.0.clone() + try_py_long_to_big_int(other)?).into_py(py))
+            Ok(PyInt(try_py_long_to_big_int(other)? + &self.0).into_py(py))
         } else {
             Ok(py.NotImplemented())
         }
@@ -318,7 +316,7 @@ impl PyInt {
     fn __rand__(&self, other: &PyAny) -> PyResult<PyObject> {
         let py = other.py();
         if other.is_instance(PyLong::type_object(py))? {
-            Ok(PyInt(self.0.clone() & try_py_long_to_big_int(other)?).into_py(py))
+            Ok(PyInt(try_py_long_to_big_int(other)? & &self.0).into_py(py))
         } else {
             Ok(py.NotImplemented())
         }
@@ -327,7 +325,7 @@ impl PyInt {
     fn __rdivmod__(&self, dividend: &PyAny) -> PyResult<PyObject> {
         let py = dividend.py();
         if dividend.is_instance(PyLong::type_object(py))? {
-            try_divmod(try_py_long_to_big_int(dividend)?, self.0.clone())
+            try_divmod(&try_py_long_to_big_int(dividend)?, &self.0)
                 .map(|(quotient, remainder)| (PyInt(quotient), PyInt(remainder)).into_py(py))
         } else {
             Ok(py.NotImplemented())
@@ -341,7 +339,7 @@ impl PyInt {
     fn __rfloordiv__(&self, dividend: &PyAny) -> PyResult<PyObject> {
         let py = dividend.py();
         if dividend.is_instance(PyLong::type_object(py))? {
-            try_floordiv(try_py_long_to_big_int(dividend)?, self.0.clone())
+            try_floordiv(&try_py_long_to_big_int(dividend)?, &self.0)
                 .map(|result| PyInt(result).into_py(py))
         } else {
             Ok(py.NotImplemented())
@@ -379,7 +377,7 @@ impl PyInt {
     fn __rmul__(&self, other: &PyAny) -> PyResult<PyObject> {
         let py = other.py();
         if other.is_instance(PyLong::type_object(py))? {
-            Ok(PyInt(self.0.clone() * try_py_long_to_big_int(other)?).into_py(py))
+            Ok(PyInt(try_py_long_to_big_int(other)? * &self.0).into_py(py))
         } else {
             Ok(py.NotImplemented())
         }
@@ -403,7 +401,7 @@ impl PyInt {
                             .checked_pow(-try_py_long_to_big_int(digits)?)
                             .unwrap_unchecked()
                     };
-                    PyInt(self.0.clone() - try_mod_to_near(self.0.clone(), ten_to_digits_power)?)
+                    PyInt(&self.0 - try_mod_to_near(&self.0, &ten_to_digits_power)?)
                 } else {
                     self.clone()
                 }
@@ -421,10 +419,10 @@ impl PyInt {
         };
         match divisor {
             Some(divisor) => match try_py_any_to_maybe_big_int(divisor)? {
-                Some(divisor) => try_pow_mod(base, self.0.clone(), divisor, py),
+                Some(divisor) => try_pow_mod(base, &self.0, divisor, py),
                 None => Ok(py.NotImplemented()),
             },
-            None => try_pow(base, self.0.clone(), py),
+            None => try_pow(&base, &self.0, py),
         }
     }
 
@@ -451,7 +449,7 @@ impl PyInt {
     fn __rsub__(&self, minuend: &PyAny) -> PyResult<PyObject> {
         let py = minuend.py();
         if minuend.is_instance(PyLong::type_object(py))? {
-            Ok(PyInt(try_py_long_to_big_int(minuend)? - self.0.clone()).into_py(py))
+            Ok(PyInt(try_py_long_to_big_int(minuend)? - &self.0).into_py(py))
         } else {
             Ok(py.NotImplemented())
         }
@@ -492,7 +490,7 @@ impl PyInt {
     fn __sub__(&self, subtrahend: &PyAny) -> PyResult<PyObject> {
         let py = subtrahend.py();
         match try_py_any_to_maybe_big_int(subtrahend)? {
-            Some(subtrahend) => Ok(PyInt(self.0.clone() - subtrahend).into_py(py)),
+            Some(subtrahend) => Ok(PyInt(&self.0 - subtrahend).into_py(py)),
             None => Ok(py.NotImplemented()),
         }
     }
@@ -545,17 +543,17 @@ fn try_divmod<
     match dividend.checked_div_rem_euclid(divisor) {
         Some((quotient, remainder)) => Ok((quotient, remainder)),
         None => Err(PyZeroDivisionError::new_err(
-            UNDEFINED_DIVISION_ERROR_MESSAGE,
+            big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
         )),
     }
 }
 
 #[inline]
-fn try_floordiv(dividend: BigInt, divisor: BigInt) -> PyResult<BigInt> {
+fn try_floordiv(dividend: &BigInt, divisor: &BigInt) -> PyResult<BigInt> {
     match dividend.checked_div_euclid(divisor) {
         Some(result) => Ok(result),
         None => Err(PyZeroDivisionError::new_err(
-            UNDEFINED_DIVISION_ERROR_MESSAGE,
+            big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
         )),
     }
 }
@@ -563,9 +561,9 @@ fn try_floordiv(dividend: BigInt, divisor: BigInt) -> PyResult<BigInt> {
 #[inline]
 fn try_lshift(base: BigInt, shift: BigInt) -> PyResult<BigInt> {
     base.checked_shl(shift).map_err(|reason| match reason {
-        big_int::LeftShiftError::NegativeShift => PyValueError::new_err(reason.to_string()),
-        big_int::LeftShiftError::OutOfMemory => PyMemoryError::new_err(reason.to_string()),
-        big_int::LeftShiftError::TooLarge => PyOverflowError::new_err(reason.to_string()),
+        big_int::ShlError::NegativeShift => PyValueError::new_err(reason.to_string()),
+        big_int::ShlError::OutOfMemory => PyMemoryError::new_err(reason.to_string()),
+        big_int::ShlError::TooLarge => PyOverflowError::new_err(reason.to_string()),
     })
 }
 
@@ -574,33 +572,33 @@ fn try_mod(dividend: BigInt, divisor: BigInt) -> PyResult<BigInt> {
     match dividend.checked_rem_euclid(divisor) {
         Some(result) => Ok(result),
         None => Err(PyZeroDivisionError::new_err(
-            UNDEFINED_DIVISION_ERROR_MESSAGE,
+            big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
         )),
     }
 }
 
 #[inline]
-fn try_mod_to_near(dividend: BigInt, divisor: BigInt) -> PyResult<BigInt> {
-    let (quotient, remainder) = match dividend.checked_div_rem_euclid(divisor.clone()) {
+fn try_mod_to_near(dividend: &BigInt, divisor: &BigInt) -> PyResult<BigInt> {
+    let (quotient, remainder) = match dividend.checked_div_rem_euclid(divisor) {
         Some((quotient, remainder)) => Ok((quotient, remainder)),
         None => Err(PyZeroDivisionError::new_err(
-            UNDEFINED_DIVISION_ERROR_MESSAGE,
+            big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
         )),
     }?;
     let double_remainder = remainder
         .clone()
         .checked_shl(BigInt::one())
         .map_err(|reason| match reason {
-            big_int::LeftShiftError::NegativeShift => PyValueError::new_err(reason.to_string()),
-            big_int::LeftShiftError::OutOfMemory => PyMemoryError::new_err(reason.to_string()),
-            big_int::LeftShiftError::TooLarge => PyOverflowError::new_err(reason.to_string()),
+            big_int::ShlError::NegativeShift => PyValueError::new_err(reason.to_string()),
+            big_int::ShlError::OutOfMemory => PyMemoryError::new_err(reason.to_string()),
+            big_int::ShlError::TooLarge => PyOverflowError::new_err(reason.to_string()),
         })?;
     let greater_than_half = if divisor.is_positive() {
-        double_remainder > divisor
+        &double_remainder > divisor
     } else {
-        double_remainder < divisor
+        &double_remainder < divisor
     };
-    let exactly_half = double_remainder == divisor;
+    let exactly_half = &double_remainder == divisor;
     Ok(
         if greater_than_half || (exactly_half && quotient.is_odd()) {
             remainder - divisor
@@ -611,13 +609,14 @@ fn try_mod_to_near(dividend: BigInt, divisor: BigInt) -> PyResult<BigInt> {
 }
 
 #[inline]
-fn try_pow(base: BigInt, exponent: BigInt, py: Python) -> PyResult<PyObject> {
+fn try_pow(base: &BigInt, exponent: &BigInt, py: Python) -> PyResult<PyObject> {
     if exponent.is_negative() {
-        match unsafe { Fraction::new(base, BigInt::one()).unwrap_unchecked() }.checked_pow(exponent)
+        match unsafe { Fraction::new(base.clone(), BigInt::one()).unwrap_unchecked() }
+            .checked_pow(exponent.clone())
         {
             Some(power) => Ok(PyFraction(power).into_py(py)),
             None => Err(PyZeroDivisionError::new_err(
-                UNDEFINED_DIVISION_ERROR_MESSAGE,
+                big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
             )),
         }
     } else {
@@ -626,22 +625,27 @@ fn try_pow(base: BigInt, exponent: BigInt, py: Python) -> PyResult<PyObject> {
 }
 
 #[inline]
-fn try_pow_mod(base: BigInt, exponent: BigInt, divisor: BigInt, py: Python) -> PyResult<PyObject> {
-    let is_zero_divisor = divisor.is_zero();
+fn try_pow_mod<
+    Base: CheckedPowRemEuclid<Exponent, Divisor, Output = Result<BigInt, Error>>,
+    Exponent,
+    Divisor: Zeroable,
+    Error: ToString,
+>(
+    base: Base,
+    exponent: Exponent,
+    divisor: Divisor,
+    py: Python,
+) -> PyResult<PyObject> {
     match base.checked_pow_rem_euclid(exponent, divisor) {
-        Some(remainder) => Ok(PyInt(remainder).into_py(py)),
-        None => Err(PyValueError::new_err(if is_zero_divisor {
-            "Divisor cannot be zero."
-        } else {
-            "Base is not invertible for the given divisor."
-        })),
+        Ok(remainder) => Ok(PyInt(remainder).into_py(py)),
+        Err(reason) => Err(PyValueError::new_err(reason.to_string())),
     }
 }
 
 #[inline]
 fn try_rshift(base: BigInt, shift: BigInt) -> PyResult<BigInt> {
     base.checked_shr(shift).map_err(|reason| match reason {
-        big_int::RightShiftError::NegativeShift => PyValueError::new_err(reason.to_string()),
+        big_int::ShrError::NegativeShift => PyValueError::new_err(reason.to_string()),
     })
 }
 
@@ -725,7 +729,7 @@ fn try_truediv(dividend: BigInt, divisor: BigInt) -> PyResult<Fraction> {
     match Fraction::new(dividend, divisor) {
         Some(quotient) => Ok(quotient),
         None => Err(PyZeroDivisionError::new_err(
-            UNDEFINED_DIVISION_ERROR_MESSAGE,
+            big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
         )),
     }
 }
@@ -743,7 +747,7 @@ impl PyFraction {
                     ) {
                         Some(fraction) => Ok(PyFraction(fraction)),
                         None => Err(PyZeroDivisionError::new_err(
-                            UNDEFINED_DIVISION_ERROR_MESSAGE,
+                            big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
                         )),
                     }
                 }
@@ -858,7 +862,7 @@ impl PyFraction {
             {
                 Some(quotient) => Ok(PyInt(quotient).into_py(py)),
                 None => Err(PyZeroDivisionError::new_err(
-                    UNDEFINED_DIVISION_ERROR_MESSAGE,
+                    big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
                 )),
             }
         } else {
@@ -866,7 +870,7 @@ impl PyFraction {
                 Some(divisor) => match self.0.clone().checked_div_euclid(divisor) {
                     Some(quotient) => Ok(PyInt(quotient).into_py(py)),
                     None => Err(PyZeroDivisionError::new_err(
-                        UNDEFINED_DIVISION_ERROR_MESSAGE,
+                        big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
                     )),
                 },
                 None => Ok(py.NotImplemented()),
@@ -892,7 +896,7 @@ impl PyFraction {
             {
                 Some(remainder) => Ok(PyFraction(remainder).into_py(py)),
                 None => Err(PyZeroDivisionError::new_err(
-                    UNDEFINED_DIVISION_ERROR_MESSAGE,
+                    big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
                 )),
             }
         } else {
@@ -900,7 +904,7 @@ impl PyFraction {
                 Some(divisor) => match self.0.clone().checked_rem_euclid(divisor) {
                     Some(remainder) => Ok(PyFraction(remainder).into_py(py)),
                     None => Err(PyZeroDivisionError::new_err(
-                        UNDEFINED_DIVISION_ERROR_MESSAGE,
+                        big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
                     )),
                 },
                 None => Ok(py.NotImplemented()),
@@ -934,7 +938,7 @@ impl PyFraction {
                 Some(exponent) => match self.0.clone().checked_pow(exponent) {
                     Some(power) => Ok(PyFraction(power).into_py(py)),
                     None => Err(PyZeroDivisionError::new_err(
-                        UNDEFINED_DIVISION_ERROR_MESSAGE,
+                        big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
                     )),
                 },
                 None => Ok(py.NotImplemented()),
@@ -985,7 +989,7 @@ impl PyFraction {
             Some(dividend) => match dividend.checked_div_euclid(self.0.clone()) {
                 Some(quotient) => Ok(PyInt(quotient).into_py(py)),
                 None => Err(PyZeroDivisionError::new_err(
-                    UNDEFINED_DIVISION_ERROR_MESSAGE,
+                    big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
                 )),
             },
             None => Ok(py.NotImplemented()),
@@ -998,7 +1002,7 @@ impl PyFraction {
             Some(dividend) => match dividend.checked_rem_euclid(self.0.clone()) {
                 Some(remainder) => Ok(PyFraction(remainder).into_py(py)),
                 None => Err(PyZeroDivisionError::new_err(
-                    UNDEFINED_DIVISION_ERROR_MESSAGE,
+                    big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
                 )),
             },
             None => Ok(py.NotImplemented()),
@@ -1056,7 +1060,7 @@ impl PyFraction {
             Some(dividend) => match dividend.checked_div(self.0.clone()) {
                 Some(quotient) => Ok(PyFraction(quotient).into_py(py)),
                 None => Err(PyZeroDivisionError::new_err(
-                    UNDEFINED_DIVISION_ERROR_MESSAGE,
+                    big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
                 )),
             },
             None => Ok(py.NotImplemented()),
@@ -1075,7 +1079,7 @@ impl PyFraction {
                 Ok(())
             }
             None => Err(PyZeroDivisionError::new_err(
-                UNDEFINED_DIVISION_ERROR_MESSAGE,
+                big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
             )),
         }
     }
@@ -1106,7 +1110,7 @@ impl PyFraction {
             {
                 Some(quotient) => Ok(PyFraction(quotient).into_py(py)),
                 None => Err(PyZeroDivisionError::new_err(
-                    UNDEFINED_DIVISION_ERROR_MESSAGE,
+                    big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
                 )),
             }
         } else {
@@ -1114,7 +1118,7 @@ impl PyFraction {
                 Some(divisor) => match self.0.clone().checked_div(divisor) {
                     Some(quotient) => Ok(PyFraction(quotient).into_py(py)),
                     None => Err(PyZeroDivisionError::new_err(
-                        UNDEFINED_DIVISION_ERROR_MESSAGE,
+                        big_int::UNDEFINED_DIVISION_ERROR_MESSAGE,
                     )),
                 },
                 None => Ok(py.NotImplemented()),
